#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/wait.h>
#ifdef HAS_SYSLOG
#include <syslog.h>
#endif
#include "getopt.h"
#include "lutil.h"
#include "xutil.h"
#include "ftn.h"
#include "nodelist.h"
#include "trap.h"
#include "config.h"
#include "version.h"
#ifdef QDOS
#include <fcntl.h>
#include <csrvthg.h>
#include <qdos.h>
#include <qptr.h>
#endif

#if defined(SHORT_PID_T)
#define pid_t short
#elif defined(INT_PID_T)
#define pid_t int
#endif

#ifdef QDOS
long    qlchan;
void    consetup_title();
void    (*_consetup)() = consetup_title;        /* use special window setup */
char    _PROG_NAME[] = "Pbox Mailer";           /* name in window & job table */
char    _version[]   = "1.18";                  /* version in job window */
timeout_t _endtimeout = 200;                    /* Wait 4 seconds on exit */
int pserve = -1;
long bbsline = 1;
chanid_t screen;
int quietmode = 0;
char *predial=NULL;
#endif

int master=0;
int forcedcalls=0;
char *forcedphone=NULL;
char *forcedline=NULL;
#if defined(HAS_TCP) || defined(HAS_TERM)
char *inetaddr=NULL;
#endif
#ifdef NOISEDEBUG
int junklevel=0;
#endif

int call(faddr *);
fa_list *callall(void);
int answer(char *);
void mkdirs(char*);


void usage(void)
{
#ifdef NOISEDEBUG
#if defined(HAS_TCP) || defined(HAS_TERM)
        confusage("-r<role> -j<num> -a<inetaddr> <node> ...");
#else
        confusage("-d<command> -q<quiet> -r<role> -j<num> <node> ...");
#endif
        fprintf(stderr,"-j<num>         damage every <num> byte [%d]\n",
                                                                junklevel);
#else
#if defined(HAS_TCP) || defined(HAS_TERM)
        confusage("-r<role> -a<inetaddr> <node> ...");
#else
        confusage("-d<command> -q<quiet> -r<role> <node> ...");
#endif
#endif
        fprintf(stderr,"-r 0|1          1 - master, 0 - slave   [0]\n");
        fprintf(stderr,"-q 0|1          1 - quiet,  0 - normal  [0]\n");
        fprintf(stderr,"-d<command>     pre-dial modem command\n");
        fprintf(stderr,"-n<phone>       forced phone number\n");
        fprintf(stderr,"-l<ttydevice>   forced tty device\n");
        fprintf(stderr,"-b<number>      use this bbs line\n");
#if defined(HAS_TCP) || defined(HAS_TERM)
        fprintf(stderr,"-a<inetaddr>    use TCP/IP instead of modem\n");
#endif
        fprintf(stderr,"  <node>        in domain form, e.g. f11.n22.z3\n");
        fprintf(stderr,"                or fido form,   e.g. 3:22/11\n");
        fprintf(stderr,"                (this implies master mode)\n");
        fprintf(stderr,"\n or: tsync|yoohoo|**EMSI_INQC816\n");
        fprintf(stderr,"                (this implies slave mode)\n");
}

int main(argc,argv)
int argc;
char *argv[];
{
#ifdef QDOS
        short rlen;
        char *a;
#endif
        int tmr,rdl;
        int c,uid;
        fa_list *callist=NULL,**tmpl;
        faddr *tmp;
        int rc,maxrc,callno=0,succno=0;
        char *answermode=NULL,*p;
#ifdef NEED_FORK
        pid_t child,waitret;
        int status;
#endif

#if defined(HAS_SYSLOG) && defined(CICOLOG)
        logfacility=CICOLOG;
#endif
        setmyname(argv[0]);
        catch(myname);
#ifdef NOISEDEBUG
#if defined(HAS_TCP) || defined(HAS_TERM)
        while ((c=getopt(argc,argv,"j:x:r:n:l:a:I:h")) != -1)
#else
        while ((c=getopt(argc,argv,"j:x:r:n:l:I:h")) != -1)
#endif
#else
#if defined(HAS_TCP) || defined(HAS_TERM)
        while ((c=getopt(argc,argv,"x:r:n:l:a:I:h")) != -1)
#else
        while ((c=getopt(argc,argv,"d:q:b:x:r:n:l:I:h")) != -1)
#endif
#endif
        if (confopt(c,optarg)) switch (c)
        {
#ifdef NOISEDEBUG
        case 'j':       junklevel=atoi(optarg); break;
#endif
        case 'b':       bbsline=atoi(optarg); break;
        case 'd':       predial=optarg; break;
        case 'q':       quietmode=atoi(optarg);
                        if ((quietmode != 0) && (quietmode != 1))
                        {
                                usage();
                                exit(1);
                        }
                        break;
        case 'r':       master=atoi(optarg);
                        if ((master != 0) && (master != 1))
                        {
                                usage();
                                exit(1);
                        }
                        break;
        case 'l':       forcedline=optarg; break;
#if defined(HAS_TCP) || defined(HAS_TERM)
        case 'a':       inetaddr=optarg; break;
#endif
        case 'n':       forcedphone=optarg; break;
        default:        usage(); exit(1);
        }
#ifdef QDOS
        pserve = UseSrvThg (CLNT, "pbox");
        if(quietmode)
          screen = io_open("nul",0);
        else
        {
          screen = io_open("con",0);
          consetup_title(screen, &_condetails);
        }
        (void)mt_prior(-1L,127);        /* make me VERY important */
#endif
        if (readconfig())
        {
                fprintf(stderr,"Error getting configuration, aborting\n");
                exit(1);
        }

        maxpsize=0L; /* we want classic packet names here */

        tmpl=&callist;

        while (argv[optind])
        {
                for (p=argv[optind];(*p) && (*p == '*');p++);
                if ((strcasecmp(argv[optind],"tsync") == 0) ||
                    (strcasecmp(argv[optind],"yoohoo") == 0) ||
                    (strncasecmp(p,"EMSI_",5) == 0))
                {
                        master=0;
                        answermode=argv[optind];
                        debug(1,"inbound \"%s\" mode",S(answermode));
                }
                else
                {
                        debug(1,"callist entry \"%s\"",argv[optind]);
                        if ((tmp=parsefaddr(argv[optind])))
                        {
                                *tmpl=(fa_list *)xmalloc(sizeof(fa_list));
                                (*tmpl)->next=NULL;
                                (*tmpl)->addr=tmp;
                                tmpl=&((*tmpl)->next);
                        }
                        else logerr("unrecognizable address \"%s\"",argv[optind]);
                }
                optind++;
        }

        if (callist)
        {
                master=1;
                forcedcalls=1;
        }

        /*
           The following witchkraft about uid-s is necessary to make
           access() work right.  Unforunately, access() checks the real
           uid, if ifcico is invoked with supervisor real uid (as when
           called by uugetty) it returns X_OK for the magic files that
           even do not have `x' bit set.  Therefore, `reference' magic
           file requests are taken for `execute' requests (and the
           actual execution natually fails).  Here we set real uid equal
           to effective.  If real uid is not zero, all these fails, but
           in this case it is not necessary anyway.
        */

        uid=geteuid();
        seteuid(0);
        setuid(uid);
        seteuid(uid);
        debug(14,"uid=%d, euid=%d",getuid(),geteuid());

        umask(066); /* packets may contain confidential information */

#ifndef QDOS
        p=xstrcpy(inbound);
        p=xstrcat(p,"/tmp/");
        mkdirs(p);
        free(p);
#endif

        maxrc=0;
        if (master)
        {

#ifdef NEED_FORK
                if ((child=fork()))
                {
                        if (child == -1)
                        {
                                logerr("$fork() error");
                                exit(1);
                        }
                        while (((waitret=wait(&status)) != -1) &&
                               (waitret != child))
                        {
                                logerr("wait return %d, status %d,%d",
                                        waitret,status>>8,status&0xff);
                        }
                        if (status&0xff) kill(getpid(),status&0xff);
                        else exit(status>>8);
                }
#endif

                if (callist == NULL) callist=callall();
                for (tmpl=&callist;*tmpl;tmpl=&((*tmpl)->next))
                {
                  callno++;
                  rdl = 0;
                  tmr = retry_wt;
                  while(1)
                  {
                        rc=call((*tmpl)->addr);
                        loginf("call to %s %s (rc=%d)",
                                ascfnode((*tmpl)->addr,0x1f),
                                rc?"failed":"successful",rc);
                        if (rc > maxrc) maxrc=rc;
                        if (rc == 0) succno++;
                        if (rc != 2) break;     /* 2 = dial failed, busy? */
                        if (rdl >= retry_nr) break;
                        rdl++;
                        loginf("Retry %d in %d seconds",rdl,tmr);
                        sleep(tmr);
                  }
                }
        }
        else /* slave */
        {
#ifdef QDOS
   /* Qdos passed channels are infinite timeout */
   /* and we need zero timeout on stdin and stdout, the serial line */

                debug(14,"entry flags: stdin: 0x%04x, stdout: 0x%04x",
                      fcntl(0,F_GETFL,0L),fcntl(1,F_GETFL,0L));

                fcntl(0,F_SETFL,0x0004);        /* ndelay, rdonly */
                fcntl(1,F_SETFL,0x0005);        /* ndelay, wronly */
                setbuf(stdin,NULL);
                setbuf(stdout,NULL);
                debug(14,"new   flags: stdin: 0x%04x, stdout: 0x%04x",
                      fcntl(0,F_GETFL,0L),fcntl(1,F_GETFL,0L));
                qlchan = getchid(0);
#endif
                maxrc=answer(answermode);
                callno=1;
                succno=(maxrc == 0);
        }

        if (callno)
                loginf("%d of %d calls, maxrc=%d",succno,callno,maxrc);
#ifdef QDOS
        _endmsg = NULL; /* we dont want to wait on exit now */
        FreeSrvThg (CLNT, "pbox");
#endif
        return maxrc;
}
